#include "cs.h"
/* create a Householder reflection [v,beta,s]=house(x), overwrite x with v,
 * where (I-beta*v*v')*x = s*e1 and e1 = [1 0 ... 0]'.
 * Note that this CXSparse version is different than CSparse. */
CS_ENTRY cs_house (CS_ENTRY *x, CS_ENTRY *beta, CS_INT n)
{
    CS_ENTRY s = 0 ;
    CS_INT i ;
    if (!x || !beta) return (-1) ;	    /* check inputs */
    for (i = 0 ; i < n ; i++) s += x [i] * CS_CONJ (x [i]) ;
    s = sqrt (s) ;
    if (s == 0)
    {
	(*beta) = 0 ;
	x [0] = 1 ;
    }
    else
    {
	if (x [0] != 0)
	{
	    s *= x [0] / CS_ABS (x [0]) ;
	}
	s = -s ;
	x [0] -= s ;
	(*beta) = -1. / (CS_CONJ (s) * x [0]) ;
    }
    return (s) ;
}
